import abc
import random
from typing import Optional, List, Union, Iterable


class Distribution(abc.ABC):

    @abc.abstractmethod
    def generate(self, generator: random.Random) -> float:
        pass


class UniformDistribution(Distribution):
    """
    Uniform distribution given lower and upper limits
    """

    def __init__(self, lower: float, upper: float):
        self.__lower = lower
        self.__upper = upper

    def generate(self, generator: random.Random) -> float:
        return generator.uniform(self.__lower, self.__upper)


class NormalDistribution(Distribution):
    """
    Normal distribution given mean and standard deviation
    """

    def __init__(self, mean, standard_deviation):
        self.__mean = mean
        self.__standard_deviation = standard_deviation

    def generate(self, generator: random.Random) -> float:
        return generator.gauss(self.__mean, self.__standard_deviation)


class TriangleDistribution(Distribution):
    """
    Triangle distribution with provided low, high and mode
    """

    def __init__(self, lower: float, upper: float, mode: float):
        self.__low = lower
        self.__high = upper
        self.__mode = mode

    def generate(self, generator: random.Random) -> float:
        return generator.triangular(self.__low, self.__high, self.__mode)


class ConstantValue(Distribution):
    """
    Allows users to specify fixed starting parameters
    """

    def __init__(self, value: float):
        self.__value = value

    def generate(self, generator: random.Random) -> float:
        return self.__value


class Sampling(abc.ABC):

    def generate_all_samples(self, number_of_samples: int, sample_size: int) -> List[float]:
        samples: List[Union[float, None]] = [None] * number_of_samples * sample_size
        for sample_index in range(number_of_samples):
            samples[sample_index * sample_size: (sample_index + 1) * sample_size] = self.generate_sample(
                sample_size=sample_size)

        return samples

    @abc.abstractmethod
    def generate_sample(self, sample_size: int) -> Iterable[float]:
        pass


class DefaultSampling(Sampling):

    def __init__(self,
                 parameter_lower_bounds: List[float],
                 parameter_upper_bounds: List[float],
                 sample_properties: Optional[List[Distribution]],
                 seed: int):

        self.__generator = random.Random(seed)

        if type(sample_properties) is list:
            self.__sample_properties = sample_properties
        else:
            default_sample_properties = [UniformDistribution(lower=lower, upper=upper) for lower, upper in
                                         zip(parameter_lower_bounds, parameter_upper_bounds)]
            self.__sample_properties = default_sample_properties

    def generate_sample(self, sample_size: int) -> Iterable[float]:
        result = [0.0] * sample_size
        for variable_index, distribution in enumerate(self.__sample_properties):
            result[variable_index] = distribution.generate(generator=self.__generator)
        return result



# This is a dummy module to hold resources that can be shared across processes
# It will be reloaded in each run of pygosolnp

par_lower_limit = None
par_upper_limit = None
eq_values = None
ineq_lower_bounds = None
ineq_upper_bounds = None
pysolnp_property_keys = None
pysolnp_property_values = None
pygosolnp_property_keys = None
pygosolnp_property_values = None
parameter_guesses = None

obj_func = None
eq_func = None
ineq_func = None

pysolnp_delta = None
pysolnp_rho = None
pysolnp_max_major_iter = None
pysolnp_max_minor_iter = None
pysolnp_tolerance = None
pysolnp_debug = None
evaluation_type = None
number_of_parameters = None
eval_results = None
restart_results = None
restart_convergence = None



############################
# This example shows how to make your own custom distributions for use with pygosolnp
############################

from itertools import chain
from typing import List

# Scikit-optimize has a wide range of useful sampling functions, see below link
# https://scikit-optimize.github.io/stable/auto_examples/sampler/initial-sampling-method.html
import skopt

import pygosolnp


# The Sampling class is an abstract class that can be inherited and customized as you please
class GridSampling(pygosolnp.sampling.Sampling):

    def __init__(self,
                 parameter_lower_bounds: List[float],
                 parameter_upper_bounds: List[float],
                 seed):
        self.__space = skopt.space.Space(dimensions=zip(parameter_lower_bounds, parameter_upper_bounds))
        self.__seed = seed

    def generate_all_samples(self, number_of_samples: int, sample_size: int) -> List[float]:
        # Overwrite this function to define the behavior when generating starting guesses for all samples
        # By default it calls `generate_sample` number_of_samples times, however we customize it here
        grid = skopt.sampler.Grid()
        grid_values = grid.generate(dimensions=self.__space.dimensions,
                                    n_samples=number_of_samples,
                                    random_state=self.__seed)
        return list(chain.from_iterable(grid_values))

    def generate_sample(self, sample_size: int) -> List[float]:
        # This function is abstract in the base class
        # Not needed since we are generating a grid for all samples, so overwrite it with pass
        pass


# The Permutation Function has unique solution f(x) = 0 when x_i = i
def permutation_function(data):
    n = 4
    b = 0.5

    result1 = 0
    for index1 in range(1, n + 1):
        result2 = 0
        for index2 in range(1, n + 1):
            result2 += ((pow(index2, index1) + b) * (pow(data[index2 - 1] / index2, index1) - 1))
        result1 += pow(result2, 2)
    return result1


parameter_lower_bounds = [-4.0] * 4
parameter_upper_bounds = [4.0] * 4

if __name__ == '__main__':
    # Instantiate sampling object
    sampling = GridSampling(
        parameter_lower_bounds=parameter_lower_bounds,
        parameter_upper_bounds=parameter_upper_bounds,
        seed=92)

    # Note that the seed variable to pygosolnp.solve is ignored due to the custom sampling
    results = pygosolnp.solve(
        obj_func=permutation_function,
        par_lower_limit=parameter_lower_bounds,
        par_upper_limit=parameter_upper_bounds,
        number_of_restarts=6,
        number_of_simulations=2000,
        pysolnp_max_major_iter=25,
        pysolnp_tolerance=1E-9,
        start_guess_sampling=sampling)

    print(results.best_solution)

# Best solution: [0.6222222222222218, 2.2222222222222223, 3.822222222222222, 3.2888888888888888]
# Objective function value: 9.91928145483169
# Not perfect, but much better than truncated normal!



from enum import Enum
from multiprocessing import Array
from typing import Callable, Optional, Union, List

from pygosolnp.sampling import Distribution, DefaultSampling


class EvaluationType(Enum):
    OBJECTIVE_FUNC_EXCLUDE_INEQ = 1  # Exclude any guesses that violate the objective function and the evaluate with objective function
    PENALTY_BARRIER_FUNCTION = 2  # Use the penalty barrier function to evaluate objective function


class ProblemModel:

    def __init__(self,
                 obj_func: Callable,
                 par_lower_limit: List[float],
                 par_upper_limit: List[float],
                 number_of_restarts: int,
                 number_of_simulations: int,
                 eq_func: Optional[Callable] = None,
                 eq_values: Optional[List[float]] = None,
                 ineq_func: Optional[Callable] = None,
                 ineq_lower_bounds: Optional[List[float]] = None,
                 ineq_upper_bounds: Optional[List[float]] = None,
                 rho: float = 1.0,
                 max_major_iter: int = 10,
                 max_minor_iter: int = 10,
                 delta: float = 1e-05,
                 tolerance: float = 0.0001,
                 debug: bool = False,
                 number_of_processes: Optional[int] = 5,
                 start_guess_sampling: Union[None, List[Distribution], DefaultSampling] = None,
                 evaluation_type: Union[EvaluationType, int] = EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ):
        self.__obj_func = obj_func
        self.__par_lower_limit = par_lower_limit
        self.__par_upper_limit = par_upper_limit
        self.__number_of_restarts = number_of_restarts
        self.__number_of_simulations = number_of_simulations
        self.__eq_func = eq_func
        self.__eq_values = eq_values
        self.__ineq_func = ineq_func
        self.__ineq_lower_bounds = ineq_lower_bounds
        self.__ineq_upper_bounds = ineq_upper_bounds
        self.__rho = rho
        self.__max_major_iter = max_major_iter
        self.__max_minor_iter = max_minor_iter
        self.__delta = delta
        self.__tolerance = tolerance
        self.__debug = debug
        self.__number_of_processes = number_of_processes
        self.__start_guess_sampling = start_guess_sampling
        self.__evaluation_type = EvaluationType(evaluation_type)

    @property
    def obj_func(self):
        return self.__obj_func

    @property
    def number_of_restarts(self) -> int:
        return self.__number_of_restarts

    @property
    def number_of_evaluations(self) -> int:
        return self.__number_of_simulations

    @property
    def ineq_func(self) -> Callable:
        return self.__ineq_func

    @property
    def par_lower_limit(self) -> Union[List, Array]:
        return self.__par_lower_limit

    @property
    def par_upper_limit(self) -> Union[List, Array]:
        return self.__par_upper_limit

    @property
    def sample_size(self) -> int:
        return len(self.__par_lower_limit)

    @property
    def eq_func(self) -> Callable:
        return self.__eq_func

    @property
    def eq_values(self) -> Union[List, Array]:
        return self.__eq_values

    @property
    def ineq_lower_bounds(self) -> Union[List, Array]:
        return self.__ineq_lower_bounds

    @property
    def ineq_upper_bounds(self) -> Union[List, Array]:
        return self.__ineq_upper_bounds

    @property
    def delta(self) -> float:
        return self.__delta

    @property
    def rho(self) -> float:
        return self.__rho

    @property
    def tolerance(self) -> float:
        return self.__tolerance

    @property
    def max_major_iter(self) -> int:
        return self.__max_major_iter

    @property
    def max_minor_iter(self) -> int:
        return self.__max_minor_iter

    @property
    def debug(self) -> bool:
        return self.__debug

    @property
    def evaluation_type(self) -> EvaluationType:
        return self.__evaluation_type

    @property
    def number_of_parameters(self) -> int:
        return len(self.__par_lower_limit)

    @property
    def has_eq_bounds(self) -> bool:
        return self.__eq_func is not None and \
               self.__eq_values is not None

    @property
    def has_ineq_bounds(self) -> bool:
        return self.__ineq_func is not None and \
               self.__ineq_lower_bounds is not None and \
               self.__ineq_upper_bounds is not None

    @property
    def start_guess_sampling(self):
        return self.__start_guess_sampling

    def validate(self):
        mandatory_data = [self.__obj_func, self.__par_lower_limit, self.__par_upper_limit]
        if any(data is None for data in mandatory_data):
            raise ValueError(
                "obj_func, par_lower_limit and par_upper_limit are required for pygosolnp to function")

        if not callable(self.__obj_func):
            raise ValueError("obj_func must be callable")

        if not hasattr(self.__par_lower_limit, '__len__'):
            raise ValueError("par_lower_limit does not have a fixed length, is it an list or array?")

        if not hasattr(self.__par_upper_limit, '__len__'):
            raise ValueError("par_upper_limit does not have a fixed length, is it an list or array?")

        if len(self.__par_lower_limit) != len(self.__par_upper_limit):
            raise ValueError("par_lower_limit and par_upper_limit bounds are not of the same length")

        if hasattr(self.__start_guess_sampling, '__len__') and \
                len(self.__start_guess_sampling) != len(self.__par_lower_limit):
            raise ValueError(
                "random_number_distribution input must be either None, an array of distributions with the same length as par_lower_limit and par_upper_limit or a Sampling instance")

        eq_data = [self.__eq_func, self.__eq_values]
        if not (all(data is None for data in eq_data) or all(data is not None for data in eq_data)):
            raise ValueError(
                "For equality constrained problems, please provide both eq_func and eq_values, or alternatively set both to None if not applicable")

        if self.__eq_func is not None and not callable(self.__eq_func):
            raise ValueError("For equality constrained problems, please supply a callable eq_func")

        if self.__eq_values is not None and not hasattr(self.__eq_values, '__len__'):
            raise ValueError("For equality constrained problems, please supply an fixed-length eq_values")

        ineq_data = [self.__ineq_func, self.__ineq_lower_bounds, self.__ineq_upper_bounds]
        if not (all(data is None for data in ineq_data) or all(data is not None for data in ineq_data)):
            raise ValueError(
                "For inequality constrained problems, please provide ineq_func, ineq_lower_bound and ineq_upper_bounds, or alternatively set all to None if not applicable")

        if self.__ineq_func is not None and not callable(self.__ineq_func):
            raise ValueError("For inequality constrained problems, please supply a callable ineq_func")

        if self.__ineq_lower_bounds is not None and not hasattr(self.__ineq_lower_bounds, '__len__'):
            raise ValueError("For inequality constrained problems, please supply an fixed-length ineq_lower_bounds")

        if self.__ineq_upper_bounds is not None and not hasattr(self.__ineq_upper_bounds, '__len__'):
            raise ValueError("For inequality constrained problems, please supply an fixed-length ineq_upper_bounds")

        if (self.__ineq_lower_bounds is not None) and (self.__ineq_upper_bounds is not None) and (
                len(self.__ineq_lower_bounds) != len(self.__ineq_upper_bounds)):
            raise ValueError(
                "For inequality constrained problems, please make sure that ineq_lower_bound is of the same length as ineq_upper_bound")

        if self.__number_of_simulations < 1:
            raise ValueError("number_of_simulations needs to be a positive integer value")

        if self.__number_of_restarts < 1:
            raise ValueError("number_of_restarts needs to be a positive integer value")

        if self.__number_of_processes is not None and self.__number_of_processes < 1:
            raise ValueError(
                "number_of_processes needs to be a positive integer value and is recommended to be greater than or equal to 2")

        if type(self.__rho) is not float:
            raise ValueError("pysolnp_rho needs to be a float value")

        if type(self.__max_major_iter) is not int:
            raise ValueError("pysolnp_max_major_iter needs to be a int value")

        if self.__max_major_iter <= 0:
            raise ValueError("pysolnp_max_major_iter needs to be a positive value")

        if type(self.__max_minor_iter) is not int:
            raise ValueError("pysolnp_max_minor_iter needs to be a int value")

        if self.__max_minor_iter <= 0:
            raise ValueError("pysolnp_max_minor_iter needs to be a positive value")

        if type(self.__delta) is not float:
            raise ValueError("pysolnp_delta needs to be a float value")

        if type(self.__tolerance) is not float:
            raise ValueError("pysolnp_tolerance needs to be a float value")

        if type(self.__debug) is not bool:
            raise ValueError("debug needs to be a boolean value")




import unittest
from unittest.mock import patch

from pygosolnp.benchmarks.electron import Electron, obj_func, eq_func
from pygosolnp.pygosolnp import solve, EvaluationType

from tests.mock.mock_random import MockRandom


class TestPygosolnpElectron(unittest.TestCase):
    electron = Electron(number_of_charges=25)

    def test_electron_functions(self):
        optimum = [0.61433646, 0.51219787, -0.76783671, -0.99899687, -0.17556954, -0.33268115,
                   -0.03210551, -0.76888194, -0.74949967, 0.81150788, -0.72320933, -0.68025229,
                   0.45952132, 0.93840065, 0.68628241, 0.32797284, -0.10528078, 0.13167533,
                   0.92064585, 0.47118278, -0.16734767, -0.37218115, -0.31940340, -0.04146434,
                   0.36161351, -0.02432836, 0.64347663, 0.58093602, -0.02911027, 0.81443792,
                   0.92618849, 0.24959548, -0.40666050, 0.41394741, -0.26006846, -0.61183818,
                   0.04054085, -0.67193437, 0.33057648, 0.50458088, 0.06385934, -0.94738933,
                   -0.60190580, -0.35666799, -0.86096396, -0.45402367, -0.04542384, -0.88241957,
                   0.65598514, 0.92970571, -0.78866904, -0.56885074, -0.27007430, -0.03402711,
                   -0.55305172, 0.17747712, -0.96781782, 0.49340430, 0.51662151, 0.52327751,
                   -0.32034716, -0.73185604, -0.58081353, -0.10061517, 0.52384596, 0.94252629,
                   -0.30227374, 0.78763635, -0.15874245, 0.19164511, -0.87513271, 0.92704793,
                   0.34542317, 0.75363400, 0.06987819]

        objective_function = lambda x: self.electron.objective_function(x)
        equality_function = lambda x: self.electron.equality_function(x)
        equality_bounds = self.electron.equality_constraint_bounds
        upper_bounds = self.electron.parameter_upper_bound
        lower_bounds = self.electron.parameter_lower_bound

        objective_function_value = objective_function(optimum)
        equality_function_value = equality_function(optimum)

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        self.assertAlmostEqual(objective_function_value, 243.8128, 3)

        for index, value in enumerate(equality_function_value):
            self.assertAlmostEqual(value, equality_bounds[index], 6)

    @patch(target="random.Random", new=MockRandom)
    def test_electron_optimization_exclude_inequalities_single_process(self):
        objective_function = lambda x: self.electron.objective_function(x)
        equality_function = lambda x: self.electron.equality_function(x)
        equality_bounds = self.electron.equality_constraint_bounds
        upper_bounds = self.electron.parameter_upper_bound
        lower_bounds = self.electron.parameter_lower_bound

        results = solve(obj_func=objective_function,
                        eq_func=equality_function,
                        eq_values=equality_bounds,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=20,
                        number_of_simulations=20000,
                        number_of_processes=None,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ,
                        debug=False)

        optimum = results.best_solution.parameters
        equality_function_value = equality_function(optimum)

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        for index, value in enumerate(equality_function_value):
            self.assertAlmostEqual(value, equality_bounds[index], 3)

        self.assertLessEqual(results.best_solution.obj_value, 244)

    @patch(target="random.Random", new=MockRandom)
    def test_electron_optimization_penalty_barrier_function_single_process(self):
        objective_function = lambda x: self.electron.objective_function(x)
        equality_function = lambda x: self.electron.equality_function(x)
        equality_bounds = self.electron.equality_constraint_bounds
        upper_bounds = self.electron.parameter_upper_bound
        lower_bounds = self.electron.parameter_lower_bound

        results = solve(obj_func=objective_function,
                        eq_func=equality_function,
                        eq_values=equality_bounds,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=2,
                        number_of_simulations=20000,
                        number_of_processes=None,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.PENALTY_BARRIER_FUNCTION,
                        debug=False)

        optimum = results.best_solution.parameters
        equality_function_value = equality_function(optimum)

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        for index, value in enumerate(equality_function_value):
            self.assertAlmostEqual(value, equality_bounds[index], 3)

        self.assertLessEqual(results.best_solution.obj_value, 244)

    @patch(target="random.Random", new=MockRandom)
    def test_electron_optimization_exclude_inequalities_multiple_processes(self):
        equality_bounds = self.electron.equality_constraint_bounds
        upper_bounds = self.electron.parameter_upper_bound
        lower_bounds = self.electron.parameter_lower_bound

        results = solve(obj_func=obj_func,
                        eq_func=eq_func,
                        eq_values=equality_bounds,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=20,
                        number_of_simulations=20000,
                        number_of_processes=4,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ,
                        debug=False)

        optimum = results.best_solution.parameters
        equality_function_value = eq_func(optimum)

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        for index, value in enumerate(equality_function_value):
            self.assertAlmostEqual(value, equality_bounds[index], 3)

        self.assertLessEqual(results.best_solution.obj_value, 244)

    @patch(target="random.Random", new=MockRandom)
    def test_electron_optimization_penalty_barrier_function_multiple_processes(self):
        equality_bounds = self.electron.equality_constraint_bounds
        upper_bounds = self.electron.parameter_upper_bound
        lower_bounds = self.electron.parameter_lower_bound

        results = solve(obj_func=obj_func,
                        eq_func=eq_func,
                        eq_values=equality_bounds,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=2,
                        number_of_simulations=20000,
                        number_of_processes=None,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.PENALTY_BARRIER_FUNCTION,
                        debug=False)

        optimum = results.best_solution.parameters
        equality_function_value = eq_func(optimum)

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        for index, value in enumerate(equality_function_value):
            self.assertAlmostEqual(value, equality_bounds[index], 3)

        self.assertLessEqual(results.best_solution.obj_value, 244)



import json
import os
import re
from random import Random


class MockRandom(Random):
    """
    This class uses the logic of random and only replacs the uniform normal generation logic with a pre-generated file
    The purpose is to guarantee the same random numbers accross different os, architecture and Python versions
    """
    def __init__(self, x=None):
        mock_random_number_data = MockRandom.load_random_number_data()

        if x is None:
            raise Exception(f"Mock random number generation requires seed!")

        mock_random_numbers = {data["seed"]: data["samples"] for data in mock_random_number_data}

        if x not in mock_random_numbers:
            raise Exception(f"Missing random mock data for seed {x} of type {type(x)}")

        self.mock_random_number_sequence = mock_random_numbers[x]
        self.random_number_index = 0
        super().__init__(x)

    @staticmethod
    def load_random_number_data():
        """ RNG data is chunked due to GitHub limiting files to 100 MB"""
        folder_path = os.path.dirname(__file__)
        files = os.listdir(folder_path)
        mock_random_number_data_raw = ""
        regex = r"^random_numbers(\d).json$"  # File names for chunked RNG data

        sorted_files = []
        for file in files:
            match = re.match(pattern=regex, string=file)
            if match:
                index = int(match[1])
                sorted_files.append((file, index))

        sorted_files = sorted(sorted_files, key=lambda x: x[1])

        for file, index in sorted_files:
            match = re.match(pattern=regex, string=file)
            if match:
                file_path = os.path.join(folder_path, file)
                with open(file_path, "r") as input_data:
                    mock_random_number_data_raw = mock_random_number_data_raw + input_data.read()

        mock_random_number_data = json.loads(mock_random_number_data_raw)
        return mock_random_number_data

    def random(self):
        if self.random_number_index > len(self.mock_random_number_sequence) - 1:
            raise Exception(f"Missing random number for index {self.random_number_index}, sequence of length {len(self.mock_random_number_sequence)}")

        result = self.mock_random_number_sequence[self.random_number_index]
        self.random_number_index = self.random_number_index + 1

        return result



from math import sqrt


# ----------------------------------------------------------------------------------
# Some Problems in Global Optimization
# ----------------------------------------------------------------------------------

# Distribution of Electrons on a Sphere
# See https://www.mcs.anl.gov/~more/cops/cops3.pdf

def obj_func(data):
    electron = Electron(number_of_charges=25)
    return electron.objective_function(data=data)


def eq_func(data):
    electron = Electron(number_of_charges=25)
    return electron.equality_function(data=data)


class Electron:
    def __init__(self, number_of_charges):
        self.__number_of_charges = number_of_charges

    @property
    def number_of_charges(self):
        return self.__number_of_charges

    @property
    def number_of_parameters(self):
        return self.__number_of_charges * 3

    @property
    def parameter_lower_bound(self):
        return [-1] * self.number_of_parameters

    @property
    def parameter_upper_bound(self):
        return [1] * self.number_of_parameters

    def objective_function(self, data):
        n = self.number_of_charges
        x = data[0:n]
        y = data[n:2 * n]
        z = data[2 * n:3 * n]

        result = 0.0
        for i in range(0, n - 1):
            for j in range(i + 1, n):
                result += 1.0 / sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2 + (z[i] - z[j]) ** 2)

        return result

    def equality_function(self, data):
        n = self.number_of_charges
        x = data[0:n]
        y = data[n:2 * n]
        z = data[2 * n:3 * n]
        result = [None] * n
        for i in range(0, n):
            result[i] = x[i] ** 2 + y[i] ** 2 + z[i] ** 2

        return result

    @property
    def equality_constraint_bounds(self):
        return [1] * self.number_of_charges



from collections import namedtuple
from ctypes import c_int, c_double, c_bool
from functools import reduce
from heapq import nsmallest
from multiprocessing import Array, Value, Pool
from typing import Callable, Optional, Union, List

from pygosolnp.evaluation_functions import evaluate_starting_guess, pysolnp_solve, initialize_worker_process_resources
from pygosolnp.model import ProblemModel, EvaluationType
from pygosolnp.sampling import Distribution, Sampling, DefaultSampling

Result = namedtuple(typename="Result", field_names=("parameters", "obj_value", "converged"))


class Results:
    def __init__(self, results: List[Result], starting_guesses: List[float]):
        self.__results = results
        self.__starting_guesses = starting_guesses

    def __str__(self):
        return f"Results(all_results={self.all_results}, best_solution={self.best_solution}, starting_guesses={self.starting_guesses})"

    @property
    def all_results(self) -> List[Result]:
        return self.__results

    @property
    def best_solution(self) -> Optional[Result]:
        viable_solutions = [solution for solution in self.__results if solution.converged]
        if len(viable_solutions) == 0:
            return None
        return reduce(lambda first, second: first if first.obj_value < second.obj_value else second, viable_solutions)

    @property
    def starting_guesses(self) -> List[float]:
        return self.__starting_guesses


def __get_best_solutions(results: Union[List, Array], number_of_results: int):
    results = nsmallest(n=number_of_results,
                        iterable=enumerate(results),
                        key=lambda value: value[1])

    if all(result[1] == float("inf") for result in results):
        raise ValueError("Evaluation functions could not locate any successful starting guesses.")

    return results


def __debug_message_eval_functions(model: ProblemModel, eval_results: Union[Array, List]):
    number_of_failed_evaluations = sum((1 if value == float("inf") else 0) for value in eval_results)
    print(
        f"Out of {model.number_of_evaluations} evaluations {number_of_failed_evaluations} failed or returned infinity for evaluation function {model.evaluation_type.name}. Check for issues with your problem definition or try changing the evaluation function.")


def solve(obj_func: Callable,
          par_lower_limit: List[float],
          par_upper_limit: List[float],
          eq_func: Optional[Callable] = None,
          eq_values: Optional[List[float]] = None,
          ineq_func: Optional[Callable] = None,
          ineq_lower_bounds: Optional[List[float]] = None,
          ineq_upper_bounds: Optional[List[float]] = None,
          number_of_restarts: int = 1,
          number_of_simulations: int = 20000,
          number_of_processes: Optional[int] = None,
          start_guess_sampling: Union[None, List[Distribution], Sampling] = None,
          seed: Union[None, int] = None,
          evaluation_type: Union[EvaluationType, int] = EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ,
          pysolnp_rho: float = 1.0,
          pysolnp_max_major_iter: int = 10,
          pysolnp_max_minor_iter: int = 10,
          pysolnp_delta: float = 1e-05,
          pysolnp_tolerance: float = 0.0001,
          debug: bool = False) -> Results:
    # Represent the problem with the below object
    model = ProblemModel(obj_func=obj_func,
                         par_lower_limit=par_lower_limit,
                         par_upper_limit=par_upper_limit,
                         number_of_restarts=number_of_restarts,
                         number_of_simulations=number_of_simulations,
                         eq_func=eq_func,
                         eq_values=eq_values,
                         ineq_func=ineq_func,
                         ineq_lower_bounds=ineq_lower_bounds,
                         ineq_upper_bounds=ineq_upper_bounds,
                         rho=pysolnp_rho,
                         max_major_iter=pysolnp_max_major_iter,
                         max_minor_iter=pysolnp_max_minor_iter,
                         delta=pysolnp_delta,
                         tolerance=pysolnp_tolerance,
                         debug=debug,
                         number_of_processes=number_of_processes,
                         start_guess_sampling=start_guess_sampling,
                         evaluation_type=evaluation_type)

    # Validate the inputs for the problem model
    model.validate()

    if start_guess_sampling is None or type(start_guess_sampling) is list:
        # Generate samples using the DefaultSampling object
        sampling = DefaultSampling(parameter_lower_bounds=par_lower_limit,
                                   parameter_upper_bounds=par_upper_limit,
                                   sample_properties=start_guess_sampling,
                                   seed=seed)
    elif isinstance(start_guess_sampling, Sampling):
        if seed is not None and debug is True:
            print(f"Warning: Seed value {seed} ignored due to user sampling override")
        # User provided Sampling instance
        sampling = start_guess_sampling
    else:
        raise ValueError(
            f"Provided parameter start_guess_sampling was not of expected type. Expected None, List[Distribution] or Sampling.")

    parameter_guesses = sampling.generate_all_samples(
        number_of_samples=model.number_of_evaluations,
        sample_size=model.sample_size)

    if debug is True:
        if any(guess is None for guess in parameter_guesses):
            print(f"Some of the random samples provided failed to generate, is your Sampling class setup correctly?")

    if number_of_processes:
        par_lower_limit = Array(c_double, model.par_lower_limit, lock=False)
        par_upper_limit = Array(c_double, model.par_upper_limit, lock=False)
        eq_values = Array(c_double, model.eq_values, lock=False) if model.has_eq_bounds else None
        ineq_lower_bounds = Array(c_double, model.ineq_lower_bounds, lock=False) if model.has_ineq_bounds else None
        ineq_upper_bounds = Array(c_double, model.ineq_upper_bounds, lock=False) if model.has_ineq_bounds else None
        pysolnp_delta = Value(c_double, model.delta, lock=False)
        pysolnp_rho = Value(c_double, model.rho, lock=False)
        pysolnp_max_major_iter = Value(c_int, model.max_major_iter, lock=False)
        pysolnp_max_minor_iter = Value(c_int, model.max_minor_iter, lock=False)
        pysolnp_tolerance = Value(c_double, model.tolerance, lock=False)
        pysolnp_debug = Value(c_bool, model.debug, lock=False)
        evaluation_type = Value(c_int, model.evaluation_type.value, lock=False)
        number_of_parameters = Value(c_int, model.number_of_parameters, lock=False)

        parameter_guesses = Array(c_double, parameter_guesses, lock=False)
        eval_results = Array(c_double, model.number_of_evaluations)  # Results from the eval function
        restart_results = Array(c_double,
                                model.number_of_restarts * model.number_of_parameters)  # Results from pysolnp restarts
        restart_convergence = Array(c_bool, model.number_of_restarts)
        initargs = (
            obj_func,
            par_lower_limit,
            par_upper_limit,
            eq_func,
            eq_values,
            ineq_func,
            ineq_lower_bounds,
            ineq_upper_bounds,
            parameter_guesses,
            pysolnp_delta,
            pysolnp_rho,
            pysolnp_max_major_iter,
            pysolnp_max_minor_iter,
            pysolnp_tolerance,
            pysolnp_debug,
            evaluation_type,
            number_of_parameters,
            eval_results,
            restart_results,
            restart_convergence
        )
        with Pool(processes=number_of_processes,
                  initializer=initialize_worker_process_resources,
                  initargs=initargs) as pool:

            pool.map(evaluate_starting_guess, range(model.number_of_evaluations))

            if debug is True:
                __debug_message_eval_functions(model=model, eval_results=eval_results)

            best_evaluations = __get_best_solutions(results=eval_results, number_of_results=model.number_of_restarts)
            solve_guess_indices = [index for index, value in best_evaluations]
            # The found optimums are stored in parameter_guesses
            pool.starmap(pysolnp_solve, enumerate(solve_guess_indices))

    else:
        eval_results = [None] * model.number_of_evaluations
        restart_results = [None] * model.number_of_restarts * model.number_of_parameters
        restart_convergence = [None] * model.number_of_restarts

        initialize_worker_process_resources(
            obj_func=obj_func,
            par_lower_limit=model.par_lower_limit,
            par_upper_limit=model.par_upper_limit,
            eq_func=eq_func if model.has_eq_bounds else None,
            eq_values=model.eq_values if model.has_eq_bounds else None,
            ineq_func=ineq_func if model.has_ineq_bounds else None,
            ineq_lower_bounds=model.ineq_lower_bounds if model.has_ineq_bounds else None,
            ineq_upper_bounds=model.ineq_upper_bounds if model.has_ineq_bounds else None,
            parameter_guesses=parameter_guesses,
            pysolnp_delta=model.delta,
            pysolnp_rho=model.rho,
            pysolnp_max_major_iter=model.max_major_iter,
            pysolnp_max_minor_iter=model.max_minor_iter,
            pysolnp_tolerance=model.tolerance,
            pysolnp_debug=model.debug,
            evaluation_type=model.evaluation_type.value,
            number_of_parameters=model.number_of_parameters,
            eval_results=eval_results,
            restart_results=restart_results,
            restart_convergence=restart_convergence
        )

        for index in range(model.number_of_evaluations):
            evaluate_starting_guess(simulation_index=index)

        if debug is True:
            __debug_message_eval_functions(model=model, eval_results=eval_results)

        best_evaluations = __get_best_solutions(results=eval_results, number_of_results=model.number_of_restarts)
        solve_guess_indices = [index for index, value in best_evaluations]
        # The found optimums are stored in parameter_guesses
        for solve_index, guess_index in enumerate(solve_guess_indices):
            pysolnp_solve(solve_index=solve_index, guess_index=guess_index)

    # For each restart, get the resulting parameters
    solutions = [(restart_results[index * model.number_of_parameters: (index + 1) * model.number_of_parameters],
                  restart_convergence[index]) for index in range(model.number_of_restarts)]

    # Each Result represents a solution to the restart (might have not converged)
    all_results = [
        Result(parameters=solution, obj_value=obj_func(solution), converged=converged)
        for solution, converged in solutions]

    # pysolnp might have not converged for some solution, if no converging solutions exist, print an warning message.
    if len([solution for solution in all_results if solution.converged]) == 0:
        print(f"Not able to find any feasible solution in {number_of_restarts} restarts.")

    return Results(results=all_results, starting_guesses=parameter_guesses)



from random import Random

# This script generates random numbers based on seeds and stores them to random_numbers.json
# The reason for this is to provide the same RNG output for tests regardless of operating system or Python version

import json
import os
from random import Random


def generate():
    sample_data = [
        {
            "seed": 15,
            "number_of_samples": 200000 * 10
        },
        {
            "seed": 123,
            "number_of_samples": 10 * 10
        },
        {
            "seed": 442,
            "number_of_samples": 20000 * 3 * 25
        },
        {
            "seed": 443,
            "number_of_samples": 20000 * 3 * 25
        },
        {
            "seed": 1234567,
            "number_of_samples": 10000000
        }
    ]

    for sample in sample_data:
        seed = sample["seed"]
        number_of_samples = sample["number_of_samples"]
        generator = Random(seed)
        sample_numbers = []
        for index in range(number_of_samples):
            sample_numbers.append(generator.random())
        sample["samples"] = sample_numbers

    data_string = json.dumps(sample_data)
    number_of_chunks = 6
    data_length = len(data_string)
    chunk_size = data_length // number_of_chunks
    chunks = list(range(0, data_length, chunk_size))

    start_index = 0
    final_index = data_length

    start_indices = chunks
    end_indices = chunks[1:] + [final_index]
    intervals = zip(start_indices, end_indices)

    for index, interval in enumerate(intervals):
        path = os.path.join(os.path.dirname(__file__), f"random_numbers{index}.json")
        with open(path, "w") as output_data:
            output_data.write(data_string[interval[0]: interval[1]])


if __name__ == "__main__":
    generate()



import itertools
import unittest

from pygosolnp.pygosolnp import solve
from pygosolnp.sampling import NormalDistribution
from .resources import alkyla_equality_function, parameter_lower_bounds, parameter_upper_bounds, \
    alkyla_inequality_function, alkyla_objective_function, inequality_lower_bounds, inequality_upper_bounds, \
    equality_values


class TestPygosolnpInputValidation(unittest.TestCase):

    def test_bad_objective_function(self):
        # Missing mandatory data
        with self.assertRaises(ValueError):
            solve(obj_func=None,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds)

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=None,
                  par_upper_limit=parameter_upper_bounds)

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=None)

        # Non-callable objective function
        with self.assertRaises(ValueError):
            solve(obj_func={"hello": "world"},
                  par_lower_limit=[1, 2, 3],
                  par_upper_limit=parameter_upper_bounds)

        # Lower bounds missing length attribute
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=2,
                  par_upper_limit=parameter_upper_bounds)

        # Upper bounds missing length attribute
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=1515,
                  par_upper_limit=parameter_upper_bounds)

        # Different length of bounds
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=[1, 2, 3],
                  par_upper_limit=parameter_upper_bounds)

        # Random number distribution not same length as parameters
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  start_guess_sampling=[NormalDistribution(1, 0.1)])

    def test_bad_eq_constraints(self):
        # Eq function without values
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  start_guess_sampling=[NormalDistribution(1, 0.1)],
                  eq_func=alkyla_equality_function)

        # Eq values without function
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  start_guess_sampling=[NormalDistribution(1, 0.1)],
                  eq_values=equality_values)

        # Eq func not callable without function
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  start_guess_sampling=[NormalDistribution(1, 0.1)],
                  eq_func={"Not", "Callable"},
                  eq_values=equality_values)

        # Eq values not of fixed length
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  start_guess_sampling=[NormalDistribution(1, 0.1)],
                  eq_func=alkyla_equality_function,
                  eq_values=2)

    def test_bad_ineq_constraints(self):
        # Ineq func without bounds
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function)

        # Ineq lower bound without func or upper bound
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_lower_bounds=inequality_lower_bounds)

        # Ineq func not callable
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=[1, 2, 3],
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds)

        # Ineq lower bounds not of fixed size
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=itertools.cycle("abcd"),
                  ineq_upper_bounds=inequality_upper_bounds)

        # Ineq upper bounds not of fixed size
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=itertools.cycle("efgh"))

        # Ineq lower and upper bounds of different length
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=[1, 2, 3],
                  ineq_upper_bounds=inequality_upper_bounds)

    def test_bad_pygosolnp_parameters(self):
        # Needs strictly more than 0 simulations
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  number_of_simulations=0)

        # Needs strictly more than 0 restarts
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  number_of_restarts=0)

        # Needs strictly more than 0 processes
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  number_of_processes=0)

    def test_bad_pysolnp_parameters(self):
        # Various pysolnp parameters as wrong variable type
        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_tolerance="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_max_minor_iter="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_max_major_iter="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_rho="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_delta="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  debug="a")

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_max_minor_iter=0)

        with self.assertRaises(ValueError):
            solve(obj_func=alkyla_objective_function,
                  par_lower_limit=parameter_lower_bounds,
                  par_upper_limit=parameter_upper_bounds,
                  ineq_func=alkyla_inequality_function,
                  ineq_lower_bounds=inequality_lower_bounds,
                  ineq_upper_bounds=inequality_upper_bounds,
                  eq_func=alkyla_equality_function,
                  eq_values=equality_values,
                  pysolnp_max_major_iter=0)




import unittest
from unittest.mock import patch

from pygosolnp.pygosolnp import solve, EvaluationType
from pygosolnp.sampling import NormalDistribution, UniformDistribution, TriangleDistribution, ConstantValue
from tests.resources import alkyla_equality_function, parameter_lower_bounds, parameter_upper_bounds, \
    alkyla_inequality_function, alkyla_objective_function, inequality_lower_bounds, inequality_upper_bounds, \
    equality_values
from tests.mock.mock_random import MockRandom


class TestPygosolnpFeatures(unittest.TestCase):

    def test_rng_without_seed(self):
        # Note: No need to patch random since we are expecting actual random outcome in this test
        # Run this twice and assure results are different
        result1 = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds)

        result2 = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds)

        self.assertNotEqual(str(result1), str(result2))

    @patch(target="random.Random", new=MockRandom)
    def test_rng_with_seed(self):
        # RNG with seed, note that we remove the constraints to the problem to make it easier to find a solution
        seed = 15

        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        seed=seed)

        self.assertAlmostEqual(results.best_solution.obj_value, -2597.805528666477, 4)

    @patch(target="random.Random", new=MockRandom)
    def test_evaluation_type(self):
        seed = 1234567

        # Eval Type OBJECTIVE_FUNC_EXCLUDE_INEQ has a hard time locating starting points for problems with narrow inequality bounds
        # Alkyla is a bad function in this sense, but with enough luck and enough simulations we find something...
        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        eq_func=alkyla_equality_function,
                        eq_values=equality_values,
                        ineq_func=alkyla_inequality_function,
                        ineq_lower_bounds=inequality_lower_bounds,
                        ineq_upper_bounds=inequality_upper_bounds,
                        seed=seed,
                        evaluation_type=EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ,
                        number_of_simulations=100000)

        self.assertAlmostEqual(results.best_solution.obj_value, -172.64110132537394, 4)

        # Eval Type PENALTY_BARRIER_FUNCTION is better at locating starting points for problems with narrow inequality bounds
        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        eq_func=alkyla_equality_function,
                        eq_values=equality_values,
                        ineq_func=alkyla_inequality_function,
                        ineq_lower_bounds=inequality_lower_bounds,
                        ineq_upper_bounds=inequality_upper_bounds,
                        seed=seed,
                        evaluation_type=EvaluationType.PENALTY_BARRIER_FUNCTION.value)

        self.assertAlmostEqual(results.best_solution.obj_value, -170.98167226891587, 4)

    @patch(target="random.Random", new=MockRandom)
    def test_number_of_restarts(self):
        number_of_restarts = 2
        seed = 1234567
        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        number_of_restarts=number_of_restarts,
                        seed=seed)

        self.assertEqual(len(results.all_results), 2)

    @patch(target="random.Random", new=MockRandom)
    def test_number_of_simulations(self):
        number_of_simulations = 5
        seed = 1234567
        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        number_of_simulations=number_of_simulations,
                        seed=seed)

        # We have one guess for each parameter in each simulation
        expected_number_of_starting_guesses = number_of_simulations * len(parameter_lower_bounds)

        self.assertEqual(len(results.starting_guesses), expected_number_of_starting_guesses)

    @patch(target="random.Random", new=MockRandom)
    def test_distribution_settings(self):
        # Test the built-in distributions:
        ## UniformDistribution (default)
        ## TriangleDistribution
        ## ConstantValue (not really a distribution, but anyway...)
        ## NormalDistribution (not bound to the problem limits, so can cause ValueErrors)
        # For examples with truncated normal distribution or grid sampling, see the pygosolnp/python_examples/ folder.

        seed = 1234567
        random_number_distribution = [
            UniformDistribution(lower=parameter_lower_bounds[0], upper=parameter_upper_bounds[0]),
            UniformDistribution(lower=parameter_lower_bounds[1], upper=parameter_upper_bounds[1]),
            UniformDistribution(lower=parameter_lower_bounds[2], upper=parameter_upper_bounds[2]),
            UniformDistribution(lower=parameter_lower_bounds[3], upper=parameter_upper_bounds[3]),
            UniformDistribution(lower=parameter_lower_bounds[4], upper=parameter_upper_bounds[4]),
            UniformDistribution(lower=parameter_lower_bounds[5], upper=parameter_upper_bounds[5]),
            UniformDistribution(lower=parameter_lower_bounds[6], upper=parameter_upper_bounds[6]),
            TriangleDistribution(lower=parameter_lower_bounds[7],
                                 upper=parameter_upper_bounds[7],
                                 mode=(parameter_upper_bounds[7] + parameter_lower_bounds[7]) / 2),
            ConstantValue(value=3.6),  # Forced to always be 3.6
            NormalDistribution(mean=parameter_lower_bounds[9] + (parameter_upper_bounds[9] - parameter_lower_bounds[9]) / 2,
                               standard_deviation=(parameter_upper_bounds[9] - parameter_lower_bounds[9]) / 2)
        ]

        results = solve(obj_func=alkyla_objective_function,
                        par_lower_limit=parameter_lower_bounds,
                        par_upper_limit=parameter_upper_bounds,
                        eq_func=alkyla_equality_function,
                        eq_values=equality_values,
                        ineq_func=alkyla_inequality_function,
                        ineq_lower_bounds=inequality_lower_bounds,
                        ineq_upper_bounds=inequality_upper_bounds,
                        number_of_simulations=10,
                        number_of_restarts=1,
                        seed=seed,
                        evaluation_type=EvaluationType.PENALTY_BARRIER_FUNCTION,
                        start_guess_sampling=random_number_distribution)

        expected = [19.262864952192448, 3.771726139520572, 69.77797447275604, 19.068793973427482, 19.33182084216137, 89.89207643509036, 16.299077190999093, 7.656803716342179, 3.6, 141.68084366746555, 16.527810282976148, 2.8024116596314954, 113.8847499219252, 41.913648375763266, 13.653863317763928, 89.36107961371863, 31.65003670354735, 6.742247466242993, 3.6, 144.41878023451304, 9.155674825062048, 10.587247668990255, 89.6903787978092, 38.29483391669842, 8.064799536849511, 89.43255250817325, 86.41114587298019, 5.961162104738718, 3.6, 147.6264662441887, 10.067591751749159, 13.136604255021913, 73.43265943827133, 44.724776804281476, 0.7740903344904604, 88.462319551243, 50.21684783727512, 5.77672041074722, 3.6, 145.39269004124904, 9.068225746042762, 12.61405383313014, 68.82767844281483, 16.611623195615945, 9.00169037647699, 85.31447609370339, 51.318693144764346, 9.28229782560638, 3.6, 150.88899802298727, 2.853206753025248, 13.650346506177705, 12.332917949430415, 43.013069327734556, 19.840074115550717, 91.20142486328595, 79.33118000010684, 3.8956025310633953, 3.6, 153.4160006648588, 0.6434958413807812, 2.9463400863302613, 102.90608010105203, 38.77132597397568, 3.023003386685923, 85.51893369450241, 59.20378581486704, 4.856217794173671, 3.6, 159.0155564965776, 19.650096860859275, 14.557888535778906, 24.793341705879453, 28.499566032377153, 12.098785988152493, 92.47510246654802, 44.847907830159485, 7.819085343574262, 3.6, 154.12786686990134, 8.266201078057069, 10.758463796133531, 0.2800364042800263, 23.509199466708623, 6.6730557487734306, 86.50613863644213, 65.27792963956803, 10.344369491252513, 3.6, 153.0090154093769, 13.044272915070119, 3.896854881633315, 86.85241703973514, 22.201841915524042, 17.718325429729628, 89.780805965076, 64.98014550674284, 8.237553597176323, 3.6, 150.8605530994913]
        self.assertListEqual(results.starting_guesses, expected)



import unittest
from unittest.mock import patch

from pygosolnp.benchmarks.permutations import permutation_function, parameter_lower_bounds, parameter_upper_bounds
from pygosolnp.pygosolnp import solve, EvaluationType

from tests.mock.mock_random import MockRandom


class TestPygosolnpPermutation(unittest.TestCase):

    def test_permutations_ideal_solution_functions(self):
        # This is the theoretical optimum
        optimum = [1.0, 2.0, 3.0, 4.0]
        objective_function = permutation_function
        objective_function_value = objective_function(optimum)
        self.assertAlmostEqual(objective_function_value, 0.0, 5)

    def test_permutations_close_solution_functions(self):
        # This is the solution we get from RSOLNP
        optimum = [1.000085, 1.999737, 3.000211, 3.999942]
        objective_function = permutation_function
        objective_function_value = objective_function(optimum)
        self.assertAlmostEqual(objective_function_value, 1.883935e-08, 5)

    @patch(target="random.Random", new=MockRandom)
    def test_permuation_optimization_exclude_indequalities_single_process(self):
        objective_function = permutation_function
        upper_bounds = parameter_upper_bounds
        lower_bounds = parameter_lower_bounds

        results = solve(obj_func=objective_function,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=20,
                        number_of_simulations=20000,
                        number_of_processes=None,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ,
                        debug=False)

        optimum = results.best_solution.parameters

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        self.assertLessEqual(results.best_solution.obj_value, 0.5)

    @patch(target="random.Random", new=MockRandom)
    def test_permutations_optimization_penalty_barrier_function_single_process(self):
        objective_function = permutation_function
        upper_bounds = parameter_upper_bounds
        lower_bounds = parameter_lower_bounds

        results = solve(obj_func=objective_function,
                        par_lower_limit=lower_bounds,
                        par_upper_limit=upper_bounds,
                        number_of_restarts=20,
                        number_of_simulations=20000,
                        number_of_processes=None,
                        seed=443,
                        pysolnp_max_major_iter=100,
                        evaluation_type=EvaluationType.PENALTY_BARRIER_FUNCTION,
                        debug=False)

        optimum = results.best_solution.parameters

        for index, value in enumerate(optimum):
            self.assertGreaterEqual(value, lower_bounds[index])
            self.assertLessEqual(value, upper_bounds[index])

        self.assertLessEqual(results.best_solution.obj_value, 0.5)



import unittest
from unittest.mock import patch

from tests.mock.mock_random import MockRandom
from random import Random


class TestMockRandom(unittest.TestCase):

    def test_random(self):
        mock_random = MockRandom(x=442)
        first_random_number = mock_random.random()
        self.assertEqual(0.36640772799152665, first_random_number)
        random_normal_no_variance = mock_random.normalvariate(mu=1, sigma=0)
        self.assertEqual(1, random_normal_no_variance)
        random_normal = mock_random.normalvariate(mu=1, sigma=1)
        self.assertEqual(0.7481635417542626, random_normal)

    @patch(target="random.Random", new=MockRandom)
    def test_patch_random_mock(self):
        mock_random = Random(442)
        first_random_number = mock_random.random()
        self.assertEqual(0.36640772799152665, first_random_number)
        random_normal_no_variance = mock_random.normalvariate(mu=1, sigma=0)
        self.assertEqual(1, random_normal_no_variance)
        random_normal = mock_random.normalvariate(mu=1, sigma=1)
        self.assertEqual(0.7481635417542626, random_normal)



import os
import setuptools
from datetime import datetime

__version__ = "2021.5.1"

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as file:
    long_description = file.read()

setuptools.setup(
    name="pygosolnp",
    version=__version__,
    author='Krister Sune Jakobsson',
    author_email='krister.s.jakobsson@gmail.com',
    description='This provides the GOSOLNP optimizaiton method.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/KristerSJakobsson/pygosolnp',
    license='Boost Software License',
    packages=setuptools.find_packages(),
    install_requires=["pysolnp"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from ctypes import c_long, c_bool, c_double, c_int
from typing import Any

import pysolnp

from pygosolnp import resources
from pygosolnp.model import EvaluationType


def initialize_worker_process_resources(obj_func,
                                        par_lower_limit,
                                        par_upper_limit,
                                        eq_func,
                                        eq_values,
                                        ineq_func,
                                        ineq_lower_bounds,
                                        ineq_upper_bounds,
                                        parameter_guesses,
                                        pysolnp_delta,
                                        pysolnp_rho,
                                        pysolnp_max_major_iter,
                                        pysolnp_max_minor_iter,
                                        pysolnp_tolerance,
                                        pysolnp_debug,
                                        evaluation_type,
                                        number_of_parameters,
                                        eval_results,
                                        restart_results,
                                        restart_convergence):
    """
    This function is used to provide the functions in this file access to shared resources when running pygosolnp.
     In multiprocess-mode it will pass multiprocess-safe types for each process, in single-processing it will simply have regular Python types.
    :param obj_func: A pickleable (global) callback objective function
    :param par_lower_limit: An double / multiprocessing.Array (float) representing parameter lower constraints
    :param par_upper_limit: An double / multiprocessing.Array (float) representing parameter upper constraints
    :param eq_func: [Optional, default None] A pickleable (global) callback equality function
    :param eq_values: [Optional, default None] An List / multiprocessing.Array (float) representing the equality constraints
    :param ineq_func: [Optional, default None] A pickleable (global) callback inequality function
    :param ineq_lower_bounds: [Optional, default None] An List / multiprocessing.Array (float) representing the inequality lower constraints
    :param ineq_upper_bounds: [Optional, default None] An List / multiprocessing.Array (float) representing the inequality upper constraints
    :param parameter_guesses: An List / multiprocessing.Array (float) representing randomly generated starting guesses
    :param pysolnp_delta: An double / multiprocessing.Value (float) representing pyolnp delta parameter
    :param pysolnp_rho: An double / multiprocessing.Value (float) representing pyolnp rho parameter
    :param pysolnp_max_major_iter: An int / multiprocessing.Value (int) representing pyolnp max major iterations parameter
    :param pysolnp_max_minor_iter: An int  / multiprocessing.Value (int) representing pyolnp max minor iterations parameter
    :param pysolnp_tolerance: An double / multiprocessing.Value (float) representing pyolnp tolerance parameter
    :param pysolnp_debug: An bool / multiprocessing.Value (bool) representing pyolnp debug parameter
    :param evaluation_type: An int / multiprocessing.Value (int) representing the EvaluationType enum mappings for pygosolnp
    :param number_of_parameters: An int / multiprocessing.Value (int) representing the number of parameters for this problem (a.k.a len(par_lower_limit))
    :param eval_results: An List / multiprocessing.Array (float) for storing the evaluation function results
    :param restart_results: An List / multiprocessing.Array (float) for storing the pysolnp calculation parameter results
    :param restart_convergence: An List / multiprocessing.Array (bool) for storing the pysolnp calculation convergence results
    """
    resources.obj_func = obj_func
    resources.par_lower_limit = par_lower_limit
    resources.par_upper_limit = par_upper_limit
    resources.eq_func = eq_func
    resources.eq_values = eq_values
    resources.ineq_func = ineq_func
    resources.ineq_lower_bounds = ineq_lower_bounds
    resources.ineq_upper_bounds = ineq_upper_bounds
    resources.parameter_guesses = parameter_guesses
    resources.pysolnp_delta = pysolnp_delta
    resources.pysolnp_rho = pysolnp_rho
    resources.pysolnp_max_major_iter = pysolnp_max_major_iter
    resources.pysolnp_max_minor_iter = pysolnp_max_minor_iter
    resources.pysolnp_tolerance = pysolnp_tolerance
    resources.pysolnp_debug = pysolnp_debug
    resources.evaluation_type = evaluation_type
    resources.number_of_parameters = number_of_parameters
    resources.eval_results = eval_results
    resources.restart_results = restart_results
    resources.restart_convergence = restart_convergence


def __resource_value(resource: Any):
    type_of_value = type(resource)
    if type_of_value in [c_long, c_double, c_int, c_bool]:
        return resource.value
    return resource


def objective_func_exclude_ineq(variables, obj_func, eq_func, eq_values, ineq_func, ineq_lower_bounds,
                                ineq_upper_bounds):
    if ineq_func is not None:
        # Exclude any inequality violations by setting their value to infinity
        ineq_values = ineq_func(variables)
        is_outside_ineq_constraints = any(
            value < ineq_lower_bounds[index] or value > ineq_upper_bounds[index] for index, value in
            enumerate(ineq_values))
        if is_outside_ineq_constraints:
            return float("inf")

    if obj_func is None:
        raise ValueError(f"The objective function supplied could not be identified.")
    obj_value = obj_func(variables)
    return obj_value


def penalty_barrier_function(variables, obj_func, eq_func, eq_values, ineq_func, ineq_lower_bounds, ineq_upper_bounds):
    objective_result = obj_func(variables)
    if ineq_func is not None and ineq_upper_bounds is not None and ineq_lower_bounds is not None:
        def barrier_value_function(value: float):
            if value <= 0.0:
                return 0.0
            else:
                return (0.9 + value) ** 2

        inequality_values = ineq_func(variables)
        for index, value in enumerate(inequality_values):
            objective_result += 100.0 * (
                    barrier_value_function(ineq_lower_bounds[index] - value) +
                    barrier_value_function(value - ineq_upper_bounds[index])
            )

    if eq_func is not None and eq_values is not None:
        equality_values = eq_func(variables)
        objective_result += sum((value - eq_values[index]) ** 2 for index, value in enumerate(equality_values)) / 100.0

    return objective_result


def evaluate_starting_guess(simulation_index: int):
    guesses = __resource_value(resources.parameter_guesses)
    eval_type = __resource_value(resources.evaluation_type)
    number_of_parameters = __resource_value(resources.number_of_parameters)

    start_index = simulation_index * number_of_parameters
    end_index = (simulation_index + 1) * number_of_parameters
    try:
        eval_objective_function = {
            EvaluationType.OBJECTIVE_FUNC_EXCLUDE_INEQ.value: objective_func_exclude_ineq,
            EvaluationType.PENALTY_BARRIER_FUNCTION.value: penalty_barrier_function
        }

        eval_func = eval_objective_function[eval_type]
        eval_result = eval_func(variables=guesses[start_index: end_index],
                                obj_func=__resource_value(resources.obj_func),
                                eq_func=__resource_value(resources.eq_func),
                                eq_values=__resource_value(resources.eq_values),
                                ineq_func=__resource_value(resources.ineq_func),
                                ineq_lower_bounds=__resource_value(resources.ineq_lower_bounds),
                                ineq_upper_bounds=__resource_value(resources.ineq_upper_bounds))
        resources.eval_results[simulation_index] = eval_result
    except Exception as ex:
        resources.eval_results[simulation_index] = float("inf")


def pysolnp_solve(solve_index: int, guess_index: int):
    number_of_parameters = __resource_value(resources.number_of_parameters)
    debug = __resource_value(resources.pysolnp_debug)
    start_value = resources.parameter_guesses[(guess_index * number_of_parameters): (
            (guess_index + 1) * number_of_parameters)]

    try:
        solve_result: pysolnp.Result = pysolnp.solve(obj_func=__resource_value(resources.obj_func),
                                                     par_start_value=start_value,
                                                     par_lower_limit=__resource_value(resources.par_lower_limit),
                                                     par_upper_limit=__resource_value(resources.par_upper_limit),
                                                     eq_func=__resource_value(resources.eq_func),
                                                     eq_values=__resource_value(resources.eq_values),
                                                     ineq_func=__resource_value(resources.ineq_func),
                                                     ineq_lower_bounds=__resource_value(resources.ineq_lower_bounds),
                                                     ineq_upper_bounds=__resource_value(resources.ineq_upper_bounds),
                                                     rho=__resource_value(resources.pysolnp_rho),
                                                     max_major_iter=__resource_value(resources.pysolnp_max_major_iter),
                                                     max_minor_iter=__resource_value(resources.pysolnp_max_minor_iter),
                                                     delta=__resource_value(resources.pysolnp_delta),
                                                     tolerance=__resource_value(resources.pysolnp_tolerance),
                                                     debug=debug)

        resources.restart_results[(solve_index * number_of_parameters): (
                (solve_index + 1) * number_of_parameters)] = solve_result.optimum
        resources.restart_convergence[solve_index] = solve_result.converged
    except ValueError as value_error:
        if debug:
            print(f"Error happened when running pysolnp for guess with index {guess_index}, ignoring this result. Error message: {value_error}")



import unittest

from pygosolnp.evaluation_functions import penalty_barrier_function
from tests.resources import alkyla_equality_function, alkyla_inequality_function, alkyla_objective_function, \
    inequality_lower_bounds, inequality_upper_bounds, equality_values


class TestPygosolnpEvaluationFunctions(unittest.TestCase):

    def test_penalty_barrier_function(self):
        testing_point = [17.45, 12.0, 110.0, 30.0, 19.74, 89.2, 92.8, 8.0, 3.6, 155.0]

        value = penalty_barrier_function(testing_point, obj_func=alkyla_objective_function,
                                         eq_func=alkyla_equality_function, eq_values=equality_values,
                                         ineq_func=alkyla_inequality_function,
                                         ineq_lower_bounds=inequality_lower_bounds,
                                         ineq_upper_bounds=inequality_upper_bounds)
        self.assertAlmostEqual(value, 348.8093, 4)



############################
# This example shows how to run pygosolnp with Truncated Normal distribution using Numpy and Scipy
############################

from typing import List, Optional

# Numpy random has the PCG64 generator which according to some research is better than Mersenne Twister
from numpy.random import Generator, PCG64

# Note that this script depends on scipy, which is not a requirement for pygosolnp
from scipy.stats import truncnorm

import pygosolnp


# The Sampling class is an abstract class that can be inherited and customized as you please
class TruncatedNormalSampling(pygosolnp.sampling.Sampling):

    def __init__(self,
                 parameter_lower_bounds: List[float],
                 parameter_upper_bounds: List[float],
                 seed: Optional[int]):
        self.__generator = Generator(PCG64(seed))
        self.__parameter_lower_bounds = parameter_lower_bounds
        self.__parameter_upper_bounds = parameter_upper_bounds

    def generate_sample(self, sample_size: int) -> List[float]:
        # This function returns random starting values for one sample
        return truncnorm.rvs(a=self.__parameter_lower_bounds,
                             b=self.__parameter_upper_bounds,
                             size=sample_size,
                             random_state=self.__generator)


# The Permutation Function has unique solution f(x) = 0 when x_i = i
def permutation_function(data):
    n = 4
    b = 0.5

    result1 = 0
    for index1 in range(1, n + 1):
        result2 = 0
        for index2 in range(1, n + 1):
            result2 += ((pow(index2, index1) + b) * (pow(data[index2 - 1] / index2, index1) - 1))
        result1 += pow(result2, 2)
    return result1


parameter_lower_bounds = [-4.0] * 4
parameter_upper_bounds = [4.0] * 4

if __name__ == '__main__':
    # Instantiate sampling object
    sampling = TruncatedNormalSampling(
        parameter_lower_bounds=parameter_lower_bounds,
        parameter_upper_bounds=parameter_upper_bounds,
        seed=99)

    # Note that the seed variable to pygosolnp.solve is ignored due to the custom sampling
    results = pygosolnp.solve(
        obj_func=permutation_function,
        par_lower_limit=parameter_lower_bounds,
        par_upper_limit=parameter_upper_bounds,
        number_of_restarts=6,
        number_of_simulations=20000,
        pysolnp_max_major_iter=25,
        pysolnp_tolerance=1E-9,
        start_guess_sampling=sampling)

    print(results.best_solution)

# Best solution: [2.651591117309446, 1.7843343303461394, 3.8557508243271172, 2.601788248290573]
# Objective function value: 101.48726054338877
# Not very good, the truncated normal function has generated samples that are mostly close to 0
# This is not very good for the permutation function
